# -*- coding: utf-8 -*-
"""This module holds classes that can be used as data soures. Note that it is
   easy to create other data sources: A data source must be iterable and
   provide dicts that map from attribute names to attribute values.
"""

from csv import DictReader

import sys
if sys.platform.startswith('java'):
    # Jython specific code
    from pygrametl.jythonmultiprocessing import Queue, Process
else:
    from multiprocessing import Queue, Process

from Queue import Empty

__author__ = "Christian Thomsen"
__maintainer__ = "Sílex Sistemas Ltda."
__version__ = '0.0.1'
__all__ = ['CSVSource', 'SQLSource', 'BackgroundSource', 'HashJoiningSource',
           'MergeJoiningSource', 'ProcessSource', 'TransformingSource',
           'UnionSource', 'CrossTabbingSource', 'FilteringSource']


CSVSource = DictReader


class SQLSource(object):
    """A class for iterating the result set of a single SQL query."""

    def __init__(self, connection, query, names=(), initsql=None, \
                     cursorarg=None):
        """Arguments:
           - connection: the PEP 249 connection to use. NOT a ConnectionWrapper!
           - query: the query that generates the result
           - names: names of attributes in the result. If not set,
             the names from the database are used. Default: ()
           - initsql: SQL that is executed before the query. The result of this
             initsql is not returned. Default: None.
           - cursorarg: if not None, this argument is used as an argument when
             the connection's cursor method is called. Default: None.
        """
        self.connection = connection
        if cursorarg is not None:
            self.cursor = connection.cursor(cursorarg)
        else:
            self.cursor = connection.cursor()
        if initsql:
            self.cursor.execute(initsql)
        self.query = query
        self.names = names
        self.executed = False

    def __iter__(self):
        try:
            if not self.executed:
                self.cursor.execute(self.query)
                names = None
                if self.names or self.cursor.description:
                    names = self.names or \
                        [t[0] for t in self.cursor.description]
            while True:
                data = self.cursor.fetchmany(500)
                if not data:
                    break
                if not names:
                    # We do this to support cursor objects that only have
                    # a meaningful .description after data has been fetched.
                    # This is, for example, the case when using a named
                    # psycopg2 cursor.
                    names = [t[0] for t in self.cursor.description]
                if len(names) != len(data[0]):
                    raise ValueError, \
                        "Incorrect number of names provided. " + \
                        "%d given, %d needed." % (len(names), len(data[0]))
                for row in data:
                    yield dict(zip(names, row))
        finally:
            try:
                self.cursor.close()
            except Exception:
                pass


class ProcessSource(object):
    """A class for iterating another source in a separate process"""

    def __init__(self, source, batchsize=500, queuesize=20):
        """Arguments:
           - source: the source to iterate
           - batchsize: the number of rows passed from the worker process each
             time it passes on a batch of rows. Must be positive. Default: 1000
           - queuesize: the maximum number of batches that can wait in a queue
             between the processes. 0 means unlimited. Default: 100
        """
        if type(batchsize) != int or batchsize < 1:
            raise ValueError, 'batchsize must be a positive integer'
        self.__source = source
        self.__batchsize = batchsize
        self.__queue = Queue(queuesize)
        p = Process(target=self.__worker)
        p.name = "Process for ProcessSource"
        p.start()

    def __worker(self):
        batch = []    
        try:
            for row in self.__source:
                batch.append(row)
                if len(batch) == self.__batchsize:
                    self.__queue.put(batch)
                    batch = []
            # We're done. Send the batch if it has any data and a signal
            if batch:
                self.__queue.put(batch)
            self.__queue.put('STOP')
        except Exception, e:
            if batch:
                self.__queue.put(batch)
            self.__queue.put('EXCEPTION')
            self.__queue.put(e)

    def __iter__(self):
        while True:
            data = self.__queue.get()
            if data == 'STOP':
                break
            elif data == 'EXCEPTION':
                exc = self.__queue.get()
                raise exc
            # else we got a list of rows from the other process
            for row in data:
                yield row

BackgroundSource = ProcessSource # for compatability
# The old thread-based BackgroundSource has been removed and 
# replaced by ProcessSource


class HashJoiningSource(object):
    """A class for equi-joining two data sources."""

    def __init__(self, src1, key1, src2, key2):
        """Arguments:
           - src1: the first source. This source is iterated row by row.
           - key1: the attribute of the first source to use in the join
           - src2: the second source. The rows of this source are all loaded
             into memory.
           - key2: the attriubte of the second source to use in the join.
        """
        self.__hash = {}
        self.__src1 = src1
        self.__key1 = key1
        self.__src2 = src2
        self.__key2 = key2

    def __buildhash(self):
        for row in self.__src2:
            keyval = row[self.__key2]
            l = self.__hash.get(keyval, [])
            l.append(row)
            self.__hash[keyval] = l
        self.__ready = True

    def __iter__(self):
        self.__buildhash()
        for row in self.__src1:
            matches = self.__hash.get(row[self.__key1], [])
            for match in matches:
                newrow = row.copy()
                newrow.update(match)
                yield newrow


JoiningSource = HashJoiningSource # for compatability


class MergeJoiningSource(object):
    """A ckass for merge-joining two sorted data sources"""

    def __init__(self, src1, key1, src2, key2):
        """Arguments:
        - src1: a data source
        - key1: the attribute to use from src1
        - src2: a data source
        - key2: the attribute to use from src2
        """
        self.__src1 = src1
        self.__key1 = key1
        self.__src2 = src2
        self.__key2 = key2
        self.__next = None

    def __iter__(self):
        iter1 = self.__src1.__iter__()
        iter2 = self.__src2.__iter__()

        row1 = iter1.next()
        keyval1 = row1[self.__key1]
        rows2 = self.__getnextrows(iter2)
        keyval2 = rows2[0][self.__key2]

        while True: # At one point there will be a StopIteration
            if keyval1 == keyval2:
                # Output rows
                for part in rows2:
                    resrow = row1.copy()
                    resrow.update(part)
                    yield resrow
                row1 = iter1.next()
                keyval1 = row1[self.__key1]
            elif keyval1 < keyval2:
                row1 = iter1.next()
                keyval1 = row1[self.__key1]
            else: # k1 > k2
                rows2 = self.__getnextrows(iter2)
                keyval2 = rows2[0][self.__key2]

    def __getnextrows(self, iter):
        res = []
        keyval = None
        if self.__next is not None:
            res.append(self.__next)
            keyval = self.__next[self.__key2]
            self.__next = None
        while True:
            try:
                row = iter.next()
            except StopIteration:
                if res:
                    return res
                else:
                    raise
            if keyval is None:
                keyval = row[self.__key2] # for the first row in this round
            if row[self.__key2] == keyval:
                res.append(row)
            else:
                self.__next = row
                return res


class TransformingSource(object):
    "A source that applies functions to the rows from another source"

    def __init__(self, source, *transformations):
        """Arguments:
        - source: a data source
        - *transformations: the transformations to apply. Must be callables 
          of the form func(row) where row is a dict. Will be applied in the
          given order.
        """
        self.__source = source
        self.__transformations = transformations

    def __iter__(self):
        for row in self.__source:
            for func in self.__transformations:
                func(row)
            yield row


class CrossTabbingSource(object):
    "A source that produces a crosstab from another source"

    def __init__(self, source, rowvaluesatt, colvaluesatt, values,
                 aggregator=None, nonevalue=0, sortrows=False):
        """Arguments:
        - source: the data source to pull data from
        - rowvaluesatt: the name of the attribute that holds the values that 
          appear as rows in the result
        - colvaluesatt: the name of the attribute that holds the values that
          appear as columns in the result
        - values: the name of the attribute that holds the values to aggregate
        - aggregator: the aggregator to use (see pygrametl.aggregators). If not
          given, pygrametl.aggregators.Sum is used to sum the values
        - nonevalue: the value to return when there is no data to aggregate.
          Default: 0
        - sortrows: A boolean deciding if the rows should be sorted.
          Default: False
        """
        self.__source = source
        self.__rowvaluesatt = rowvaluesatt
        self.__colvaluesatt = colvaluesatt
        self.__values = values
        if aggregator is None:
            from pygrametl.aggregators import Sum
            self.__aggregator = Sum()
        else:
            self.__aggregator = aggregator
        self.__nonevalue = nonevalue
        self.__sortrows = sortrows
        self.__allcolumns = set()  
        self.__allrows = set()
        
    def __iter__(self):
        for data in self.__source: # first we iterate over all source data ...
            row = data[self.__rowvaluesatt]
            col = data[self.__colvaluesatt]
            self.__allrows.add(row)
            self.__allcolumns.add(col)
            self.__aggregator.process((row, col), data[self.__values])

        # ... and then we build result rows
        for row in (self.__sortrows and sorted(self.__allrows) \
                        or self.__allrows): 
            res = {self.__rowvaluesatt : row}
            for col in self.__allcolumns:
                res[col] = \
                    self.__aggregator.finish((row, col), self.__nonevalue)
            yield res


class FilteringSource(object):
    "A source that applies a filter to another source"

    def __init__(self, source, filter=bool):
        """Arguments:
           - source: the source to filter
           - filter: a callable f(row). If the result is a True value,
             the row is passed on. If not, the row is discarded.
             Default: bool, i.e., Python's standard boolean conversion which
             removes empty rows.
        """
        self.__source = source
        self.__filter = filter

    def __iter__(self):
        for row in self.__source:
            if self.__filter(row):
                yield row


class UnionSource(object):
    """A source to union other sources (possibly with different types of rows).
    All rows are read from the 1st source before rows are read from the 2nd
    source and so on (to interleave the rows, use a RoundRobinSource)
    """

    def __init__(self, *sources):
        "Arguments: The sources to union in the order they should be used."
        self.__sources = sources

    def __iter__(self):
        for src in self.__sources:
            for row in src:
                yield row


class RoundRobinSource(object):
    "A source that reads sets of rows from sources in round robin-fashion"

    def __init__(self, sources, batchsize=500):
        """Arguments:
        - sources: a sequence of data sources
        - batchsize: the amount of rows to read from a data source before going
          to the next data source. Must be positive (to empty a source before
          going to the next, use UnionSource)
        """
        self.__sources = [iter(src) for src in sources]
        self.__sources.reverse() # we iterate it from the back in __iter__
        if not batchsize > 0:
            raise ValueError, "batchsize must be positive"
        self.__batchsize = batchsize

    def __iter__(self):
        while self.__sources:
            for i in range(len(self.__sources)-1, -1, -1): #iterate from back
                cursrc = self.__sources[i]
                # now return up to __batchsize from cursrc
                try:
                    for n in range(self.__batchsize):
                        yield cursrc.next()
                except StopIteration:
                    # we're done with this source and can delete it since
                    # we iterate the list as we do
                    del self.__sources[i]
        raise StopIteration


class DynamicForEachSource(object):
    """A source that for each given argument creates a new source that
    will be iterated by this source.

    For example, useful for directories where a CSVSource should be created
    for each file.

    The user must provide a function that when called with a single argument,
    returns a new source to iterate. A DynamicForEachSource instance can be
    given to several ProcessSource instances.
    """
    def __init__(self, seq, callee):
        """Arguments:
        - seq: A sequence with the elements for each of which a unique source 
          must be created. The elements are given (one by one) to callee.
        - callee: A function f(e) that must accept elements as those in the seq
          argument. The function should return a source which then will be
          iterated by this source. The function is called once for every
          element in seq. 
        """
        self.__queue = Queue()  # a multiprocessing.Queue
        if not callable(callee):
            raise TypeError, 'callee must be callable'
        self.__callee = callee
        for e in seq:
            # put them in a safe queue such that this object can be used from
            # different fork'ed processes
            self.__queue.put(e)

    def __iter__(self):
        while True:
            try:
                arg = self.__queue.get(False)
                src = self.__callee(arg)
                for row in src:
                    yield row
            except Empty:
                raise StopIteration


# -*- coding: utf-8 -*-
"""This module holds a ConnectionWrapper that is used with a
   JDBC Connection. The module should only be used when running Jython.
"""

import java.sql as jdbc

from copy import copy as pcopy
from datetime import datetime
from sys import modules
from threading import Thread
from Queue import Queue

import pygrametl
from pygrametl.FIFODict import FIFODict

# NOTE: This module is made for Jython.

__author__ = "Christian Thomsen"
__maintainer__ = "Sílex Sistemas Ltda."
__version__ = '0.0.1'
__all__ = ['JDBCConnectionWrapper', 'BackgroundJDBCConnectionWrapper']


class JDBCConnectionWrapper(object):
    """Wrap a JDBC Connection.

       All Dimension and FactTable communicate with the data warehouse using 
       a ConnectionWrapper. In this way, the code for loading the DW does not
       have to care about which parameter format is used.  
       This ConnectionWrapper is a special one for JDBC in Jython.
    """

    def __init__(self, jdbcconn, stmtcachesize=20):
        """Create a ConnectionWrapper around the given JDBC connection.

           If no default ConnectionWrapper already exists, the new
           ConnectionWrapper is set to be the default ConnectionWrapper.

           Arguments:
           - jdbcconn: An open JDBC Connection (not a PEP249 Connection)
           - stmtcachesize: The maximum number of PreparedStatements kept
             open. Default: 20.
        """
        if not isinstance(jdbcconn, jdbc.Connection):
            raise TypeError, '1st argument must implement java.sql.Connection'
        if jdbcconn.isClosed():
            raise ValueError, '1st argument must be an open Connection'
        self.__jdbcconn = jdbcconn
        # Add a finalizer to __prepstmts to close PreparedStatements when
        # they are pushed out
        self.__prepstmts = FIFODict(stmtcachesize, lambda k, v: v[0].close())
        self.__resultmeta = FIFODict(stmtcachesize)
        self.__resultset = None
        self.__resultnames = None
        self.__resulttypes = None
        self.nametranslator = lambda s: s
        self.__jdbcconn.setAutoCommit(False)
        if pygrametl._defaulttargetconnection is None:
            pygrametl._defaulttargetconnection = self

    def __preparejdbcstmt(self, sql):
        # Find pyformat arguments and change them to question marks while
        # appending the attribute names to a list
        names = []
        newsql = sql
        while True:
            start = newsql.find('%(')
            if start == -1:
                break
            end = newsql.find(')s', start)
            if end == -1: 
                break
            name = newsql[start+2 : end]
            names.append(name)
            newsql = newsql.replace(newsql[start:end+2], '?', 1)

        ps = self.__jdbcconn.prepareStatement(newsql)

        # Find parameter types
        types = []
        parmeta = ps.getParameterMetaData()
        for i in range(len(names)):
            types.append(parmeta.getParameterType(i+1))

        self.__prepstmts[sql] = (ps, names, types)

    def __executejdbcstmt(self, sql, args):
        if self.__resultset:
            self.__resultset.close()

        if sql not in self.__prepstmts:
            self.__preparejdbcstmt(sql)
        (ps, names, types) = self.__prepstmts[sql]

        for pos in range(len(names)): # Not very Pythonic, but we're doing Java
            if args[names[pos]] is None:
                ps.setNull(pos + 1, types[pos])
            else:
                ps.setObject(pos + 1, args[names[pos]], types[pos])

        if ps.execute():
            self.__resultset = ps.getResultSet()
            if sql not in self.__resultmeta:
                self.__resultmeta[sql] = \
                    self.__extractresultmetadata(self.__resultset)
            (self.__resultnames, self.__resulttypes) = self.__resultmeta[sql]
        else:
            self.__resultset = None
            (self.__resultnames, self.__resulttypes) = (None, None)

    def __extractresultmetadata(self, resultset):
        # Get jdbc resultset metadata. extract names and types
        # and add it to self.__resultmeta
        meta = resultset.getMetaData()
        names = []
        types = []
        for col in range(meta.getColumnCount()):
            names.append(meta.getColumnName(col+1))
            types.append(meta.getColumnType(col+1))
        return (names, types)

    def __readresultrow(self):
        if self.__resultset is None:
            return None
        result = []
        for i in range(len(self.__resulttypes)):
            e = self.__resulttypes[i] # Not Pythonic, but we need i for JDBC
            if e in (jdbc.Types.CHAR, jdbc.Types.VARCHAR, 
                     jdbc.Types.LONGVARCHAR):
                result.append(self.__resultset.getString(i+1))
            elif e in (jdbc.Types.BIT, jdbc.Types.BOOLEAN):
                result.append(self.__resultset.getBool(i+1))
            elif e in (jdbc.Types.TINYINT, jdbc.Types.SMALLINT, 
                       jdbc.Types.INTEGER):
                result.append(self.__resultset.getInt(i+1))
            elif e in (jdbc.Types.BIGINT, ):
                result.append(self.__resultset.getLong(i+1))
            elif e in (jdbc.Types.DATE, ):
                result.append(self.__resultset.getDate(i+1))
            elif e in (jdbc.Types.TIMESTAMP, ):
                result.append(self.__resultset.getTimestamp(i+1))
            elif e in (jdbc.Types.TIME, ):
                result.append(self.__resultset.getTime(i+1))
            else:
                # Try this and hope for the best...
                result.append(self.__resultset.getString(i+1))
        return tuple(result)

    def execute(self, stmt, arguments=None, namemapping=None, ignored=None):
        """Execute a statement.

           Arguments:
           - stmt: the statement to execute
           - arguments: a mapping with the arguments (default: None)
           - namemapping: a mapping of names such that if stmt uses %(arg)s
             and namemapping[arg]=arg2, the value arguments[arg2] is used 
             instead of arguments[arg]
           - ignored: An ignored argument only present to accept the same 
             number of arguments as ConnectionWrapper.execute
        """
        if namemapping and arguments:
            arguments = pygrametl.copy(arguments, **namemapping)
        self.__executejdbcstmt(stmt, arguments)

    def executemany(self, stmt, params, ignored=None):
        """Execute a sequence of statements.

           Arguments:
           - stmt: the statement to execute
           - params: a sequence of arguments
           - ignored: An ignored argument only present to accept the same 
             number of arguments as ConnectionWrapper.executemany
        """
        for paramset in params:
            self.__executejdbcstmt(stmt, paramset)

    def rowfactory(self, names=None):
        """Return a generator object returning result rows (i.e. dicts)."""
        if names is None:
            if self.__resultnames is None:
                return
            else:
                names = [self.nametranslator(t[0]) for t in self.__resultnames]
        empty = (None, ) * len(self.__resultnames)
        while True:
            tuple = self.fetchonetuple()
            if tuple == empty:
                return
            yield dict(zip(names, tuple))

    def fetchone(self, names=None):
        """Return one result row (i.e. dict)."""
        if self.__resultset is None:
            return {}
        if names is None:
            names = [self.nametranslator(t[0]) for t in self.__resultnames]
        values = self.fetchonetuple()
        return dict(zip(names, values))

    def fetchonetuple(self):
        """Return one result tuple."""
        if self.__resultset is None:
            return ()
        if not self.__resultset.next():
            return (None, ) * len(self.__resultnames)
        else:
            return self.__readresultrow()

    def fetchmanytuples(self, cnt):
        """Return cnt result tuples."""
        if self.__resultset is None:
            return []
        empty = (None, ) * len(self.__resultnames)
        result = []
        for i in range(cnt):
            tmp = self.fetchonetuple()
            if tmp == empty:
                break
            result.append(tmp)
        return result

    def fetchalltuples(self):
        """Return all result tuples"""
        if self.__resultset is None:
            return []
        result = []
        empty = (None, ) * len(self.__resultnames)
        while True:
            tmp = self.fetchonetuple()
            if tmp == empty:
                return result
            result.append(tmp)

    def rowcount(self):
        """Not implemented. Return 0. Should return the size of the result."""
        return 0 

    def getunderlyingmodule(self):
        """Return a reference to the underlying connection's module."""
        return modules[self.__class__.__module__]

    def commit(self):
        """Commit the transaction."""
        pygrametl.endload()
        self.__jdbcconn.commit()

    def close(self):
        """Close the connection to the database,"""
        self.__jdbcconn.close()

    def rollback(self):
        """Rollback the transaction."""
        self.__jdbcconn.rollback()

    def setasdefault(self):
        """Set this ConnectionWrapper as the default connection."""
        pygrametl._defaulttargetconnection = self

    def cursor(self):
        """Not implemented for this JDBC connection wrapper!"""
        raise NotImplementedError, ".cursor() not supported"

    def resultnames(self):
        if self.__resultnames is None:
            return None
        else:
            return tuple(self.__resultnames)

# BackgroundJDBCConnectionWrapper is added for experiments. It is quite similar
# to JDBCConnectionWrapper and one of them may be removed.


class BackgroundJDBCConnectionWrapper(object):
    """Wrap a JDBC Connection and do all DB communication in the background.

       All Dimension and FactTable communicate with the data warehouse using 
       a ConnectionWrapper. In this way, the code for loading the DW does not
       have to care about which parameter format is used.
       This ConnectionWrapper is a special one for JDBC in Jython and does DB 
       communication from a Thread.
    """

    def __init__(self, jdbcconn, stmtcachesize=20):
        """Create a ConnectionWrapper around the given JDBC connection """
        self.__jdbcconn = jdbcconn
        # Add a finalizer to __prepstmts to close PreparedStatements when
        # they are pushed out
        self.__prepstmts = FIFODict(stmtcachesize, lambda k, v: v[0].close())
        self.__resultmeta = FIFODict(stmtcachesize)
        self.__resultset = None
        self.__resultnames = None
        self.__resulttypes = None
        self.nametranslator = lambda s: s
        self.__jdbcconn.setAutoCommit(False)
        self.__queue = Queue(5000)
        t = Thread(target=self.__worker)
        t.setDaemon(True)  # NB: "t.daemon = True" does NOT work...
        t.setName('BackgroundJDBCConnectionWrapper')
        t.start()

    def __worker(self):
        while True:
            (sql, args) = self.__queue.get()
            self.__executejdbcstmt(sql, args)
            self.__queue.task_done()

    def __preparejdbcstmt(self, sql):
        # Find pyformat arguments and change them to question marks while
        # appending the attribute names to a list
        names = []
        newsql = sql
        while True:
            start = newsql.find('%(')
            if start == -1:
                break
            end = newsql.find(')s', start)
            if end == -1: 
                break
            name = newsql[start+2 : end]
            names.append(name)
            newsql = newsql.replace(newsql[start:end+2], '?', 1)

        ps = self.__jdbcconn.prepareStatement(newsql)

        # Find parameter types
        types = []
        parmeta = ps.getParameterMetaData()
        for i in range(len(names)):
            types.append(parmeta.getParameterType(i+1))

        self.__prepstmts[sql] = (ps, names, types)

    def __executejdbcstmt(self, sql, args):
        if self.__resultset:
            self.__resultset.close()

        if sql not in self.__prepstmts:
            self.__preparejdbcstmt(sql)
        (ps, names, types) = self.__prepstmts[sql]

        for pos in range(len(names)): # Not very Pythonic, but we're doing Java
            if args[names[pos]] is None:
                ps.setNull(pos + 1, types[pos])
            else:
                ps.setObject(pos + 1, args[names[pos]], types[pos])

        if ps.execute():
            self.__resultset = ps.getResultSet()
            if sql not in self.__resultmeta:
                self.__resultmeta[sql] = \
                    self.__extractresultmetadata(self.__resultset)
            (self.__resultnames, self.__resulttypes) = self.__resultmeta[sql]
        else:
            self.__resultset = None
            (self.__resultnames, self.__resulttypes) = (None, None)

    def __extractresultmetadata(self, resultset):
        # Get jdbc resultset metadata. extract names and types
        # and add it to self.__resultmeta
        meta = resultset.getMetaData()
        names = []
        types = []
        for col in range(meta.getColumnCount()):
            names.append(meta.getColumnName(col+1))
            types.append(meta.getColumnType(col+1))
        return (names, types)

    def __readresultrow(self):
        if self.__resultset is None:
            return None
        result = []
        for i in range(len(self.__resulttypes)):
            e = self.__resulttypes[i] # Not Pythonic, but we need i for JDBC
            if e in (jdbc.Types.CHAR, jdbc.Types.VARCHAR, 
                     jdbc.Types.LONGVARCHAR):
                result.append(self.__resultset.getString(i+1))
            elif e in (jdbc.Types.BIT, jdbc.Types.BOOLEAN):
                result.append(self.__resultset.getBool(i+1))
            elif e in (jdbc.Types.TINYINT, jdbc.Types.SMALLINT, 
                       jdbc.Types.INTEGER):
                result.append(self.__resultset.getInt(i+1))
            elif e in (jdbc.Types.BIGINT, ):
                result.append(self.__resultset.getLong(i+1))
            elif e in (jdbc.Types.DATE, ):
                result.append(self.__resultset.getDate(i+1))
            elif e in (jdbc.Types.TIMESTAMP, ):
                result.append(self.__resultset.getTimestamp(i+1))
            elif e in (jdbc.Types.TIME, ):
                result.append(self.__resultset.getTime(i+1))
            else:
                # Try this and hope for the best...
                result.append(self.__resultset.getString(i+1))
        return tuple(result)

    def execute(self, stmt, arguments=None, namemapping=None, ignored=None):
        """Execute a statement.

           Arguments:
           - stmt: the statement to execute
           - arguments: a mapping with the arguments (default: None)
           - namemapping: a mapping of names such that if stmt uses %(arg)s
             and namemapping[arg]=arg2, the value arguments[arg2] is used 
             instead of arguments[arg]
           - ignored: An ignored argument only present to accept the same 
             number of arguments as ConnectionWrapper.execute
        """
        if namemapping and arguments:
            arguments = pygrametl.copy(arguments, **namemapping)
        else:
            arguments = pcopy(arguments)
        self.__queue.put((stmt, arguments))

    def executemany(self, stmt, params, ignored=None):
        """Execute a sequence of statements.

           Arguments:
           - stmt: the statement to execute
           - params: a sequence of arguments
           - ignored: An ignored argument only present to accept the same 
             number of arguments as ConnectionWrapper.executemany
        """
        for paramset in params:
            self.__queue.put((stmt, paramset))

    def rowfactory(self, names=None):
        """Return a generator object returning result rows (i.e. dicts)."""
        self.__queue.join()
        if names is None:
            if self.__resultnames is None:
                return
            else:
                names = [self.nametranslator(t[0]) for t in self.__resultnames]
        empty = (None, ) * len(self.__resultnames)
        while True:
            tuple = self.fetchonetuple()
            if tuple == empty:
                return
            yield dict(zip(names, tuple))

    def fetchone(self, names=None):
        """Return one result row (i.e. dict)."""
        self.__queue.join()
        if self.__resultset is None:
            return {}
        if names is None:
            names = [self.nametranslator(t[0]) for t in self.__resultnames]
        values = self.fetchonetuple()
        return dict(zip(names, values))

    def fetchonetuple(self):
        """Return one result tuple."""
        self.__queue.join()
        if self.__resultset is None:
            return ()
        if not self.__resultset.next():
            return (None, ) * len(self.__resultnames)
        else:
            return self.__readresultrow()

    def fetchmanytuples(self, cnt):
        """Return cnt result tuples."""
        self.__queue.join()
        if self.__resultset is None:
            return []
        empty = (None, ) * len(self.__resultnames)
        result = []
        for i in range(cnt):
            tmp = self.fetchonetuple()
            if tmp == empty:
                break
            result.append(tmp)
        return result

    def fetchalltuples(self):
        """Return all result tuples"""
        self.__queue.join()
        if self.__resultset is None:
            return []
        result = []
        empty = (None, ) * len(self.__resultnames)
        while True:
            tmp = self.fetchonetuple()
            if tmp == empty:
                return result
            result.append(tmp)

    def rowcount(self):
        """Not implemented. Return 0. Should return the size of the result."""
        return 0 

    def getunderlyingmodule(self):
        """Return a reference to the underlying connection's module."""
        return modules[self.__class__.__module__]

    def commit(self):
        """Commit the transaction."""
        pygrametl.endload()
        self.__queue.join()
        self.__jdbcconn.commit()

    def close(self):
        """Close the connection to the database,"""
        self.__queue.join()
        self.__jdbcconn.close()

    def rollback(self):
        """Rollback the transaction."""
        self.__queue.join()
        self.__jdbcconn.rollback()

    def setasdefault(self):
        """Set this ConnectionWrapper as the default connection."""
        pygrametl._defaulttargetconnection = self

    def cursor(self):
        """Not implemented for this JDBC connection wrapper!"""
        raise NotImplementedError, ".cursor() not supported"

    def resultnames(self):
        self.__queue.join()
        if self.__resultnames is None:
            return None
        else:
            return tuple(self.__resultnames)


def Date(year, month, day):
    date = '%s-%s-%s' % \
        (str(year).zfill(4), str(month).zfill(2), str(day).zfill(2)) 
    return jdbc.Date.valueOf(date)


def Timestamp(year, month, day, hour, minute, second):
    date = '%s-%s-%s %s:%s:%s' % \
        (str(year).zfill(4), str(month).zfill(2), str(day).zfill(2),
         str(hour).zfill(2), str(minute).zfill(2), str(second).zfill(2))
    return jdbc.Timestamp.valueOf(date)


# -*- coding: utf-8 -*-
"""
  A simple mapping between keys and values, but with a limited capacity. When
  the max. capacity is reached, the first inserted key/value pair is deleted
"""

from collections import deque

__author__ = "Christian Thomsen"
__maintainer__ = "Sílex Sistemas Ltda."
__version__ = '0.0.1'
__all__ = ['FIFODict']


class FIFODict:

    """
    A simple FIFO mapping between keys and values.
    When the max. capacity is reached, the key/value pair that has been in
    the dict the longest time is removed.
    """

    def __init__(self, size, finalizer=None):
        """Create a FIFODict with the given maximum size. 

           The argument size determines the maximum size of the dict. 
           If finalizer is given, it must be a callable f(key, value).
           It is then called, when a item is removed due to the size of the
           dict reaching the maximum (finalizer is NOT called when an item 
           is explicitly deleted with del d[key] or when the dict is cleared.
        """
        if not type(size) == type(0):
            raise TypeError, "size must be an int"
        if not size > 0:
            raise ValueError, "size must be positive"
        if finalizer is not None and not callable(finalizer):
            raise TypeError, "finalizer must be None or a callable"

        self.__size = size
        self.__data = {}
        self.__order = deque()
        self.__finalizer = finalizer

    def add(self, key, val):
        """Add a key/value pair to the dict. 

           If a pair p with the same key already exists, p is replaced by the 
           new pair n, but n gets p's position in the FIFO dict and is deleted
           when the old pair p would have been deleted. When the maximum 
           capacity is reached, the pair with the oldest key is deleted 
           from the dict. 

           The argument key is the key and the argument val is the value."""
        if key in self.__data:
            self.__data[key] = val # Replace old value
        elif len(self.__order) < self.__size:
            # The dict is not full yet. Just add the new pair.
            self.__order.append(key)
            self.__data[key] = val
        else:
            # The dict is full. We have to delete the oldest item first.
            delKey = self.__order.popleft()
            if self.__finalizer:
                self.__finalizer(delKey, self.__data[delKey])
            del self.__data[delKey]
            self.__order.append(key)
            self.__data[key] = val
                
    def get(self, key, default=None):
        """Find and return the element a given key maps to.

        Look for the given key in the dict and return the associated value 
        if found. If not found, the value of default is returned."""
        return self.__data.get(key, default)

    def clear(self):
        """Delete all key/value pairs from the dict"""
        self.__data = {}
        self.__order = []

    def __setitem__(self, key, item):
        self.add(key, item)

    def __getitem__(self, key):
        return self.__data[key]

    def __len__(self):
        return len(self.__data)

    def __str__(self):
        allitems = []
        for key in self.__order:
            val = self.__data[key]
            item = "%s: %s" % (str(key), str(val))
            allitems.append(item)
        return "{%s}" % ", ".join(allitems)

    def __contains__(self, item):
        return (item in self.__data)

    def __delitem__(self, item):
        if item not in self.__data:
            raise KeyError, item

        del self.__data[item]
        self.__order.remove(item)

    def __iter__(self):
        for k in self.__order:
            yield k


# -*- coding: utf-8 -*-
from distutils.core import setup

setup(
        name            = 'pygrametlsilex',
        version         = '0.0.1',
        author          = 'rvnovaes',
        author_email    = 'roberto.novaes@silexsistemas.com.br',
        url             = 'https://bitbucket.org/silex/pygrametlsilex',
        description     = 'ETL (Extract, Transform, Load) written in Python.',
        license         = 'BSD',
        packages        = ['pygrametlsilex'],
        classifiers     = ['Development Status :: 3 - Alpha',
                           'Environment :: Console',
                           'Intended Audience :: Developers',
                           'License :: OSI Approved :: BSD License',
                           'Operating System :: OS Independent',
                           'Programming Language :: Python',
                           'Programming Language :: Python :: 2',
                           'Programming Language :: Python :: 2.7',
                           'Topic :: Software Development :: Libraries :: Python Modules',
                           'Topic :: Database',
                           'Topic :: Utilities'],
        )


"""This module contains classes for making "steps" in an ETL flow.
   Steps can be connected such that a row flows from step to step and
   each step does something with the row.
"""

import pygrametl

__author__ = "Christian Thomsen"
__maintainer__ = "Sílex Sistemas Ltda."
__version__ = '0.0.1'
__all__ = ['Step', 'SourceStep', 'MappingStep', 'ValueMappingStep',
           'PrintStep', 'DimensionStep', 'SCDimensionStep', 'RenamingStep', 
           'GarbageStep', 'ConditionalStep', 'CopyStep',
           'connectsteps']


def connectsteps(*steps):
    """Set a.next = b, b.next = c, etc. when given the steps a, b, c, ..."""
    for i in range(len(steps) - 1):
        steps[i].next = steps[i+1]


class Step(object):
    """The basic class for steps in an ETL flow."""

    __steps = {}

    def __init__(self, worker=None, next=None, name=None):
        """Arguments:
           - worker: A function f(row) that performs the Step's operation.
             If None, self.defaultworker is used. Default: None
           - next: The default next step to use. This should be 1) an instance
             of a Step, 2) the name of a Step, or 3) None.
             If if is a name, the next step will be looked up dynamically
             each time. If it is None, no default step will exist and rows
             will not be passed on. Default: None
           - name: A name for the Step instance. This is used when another
             Step (implicitly or explicitly) passes on rows. If two instanes
             have the same name, the name is mapped to the instance that was
             created the latest. Default: None
        """

        if name is not None:
            self.__class__.__steps[name] = self
        self.__name = name
        self.__redirected = False
        self.__row = None
        self.worker = (worker or self.defaultworker)
        self.next = next
 
    def process(self, row):
        """Perform the Step's operation on the given row.

           If the row is not explic
           itly redirected (see _redirect), it will
           be passed on the the next step if this has been set.
        """
        self.__redirected = False
        self.__row = row
        self.worker(row)
        self.__row = None
        if self.next is None or self.__redirected:
            return
        self._inject(row, self.next)

    def _redirect(self, target):
        """Redirect the current row to the given target.

           The target is either an instance of Step or the name of a Step
           instance.
        """
        self.__redirected = True
        self._inject(self.__row, target)

    def _inject(self, row, target=None):
        """Give a row to another Step before the current row is passed on.

           The target is either 1) an instance of Step, 2) the name of a Step
           instance, or 3) None. If None, the next default Step is used
           and must be defined.
        """
        if target is None:
            target = self.next

        if isinstance(target, Step):
            target.process(row)
        else:
            self.__class__.__steps[target].process(row)

    def __call__(self, row):
        self.process(row)

    def name(self):
        """Return the name of the Step instance"""
        return self.__name

    @classmethod
    def getstep(cls, name):
        """Return the Step instance with the given name"""
        return cls.__steps.get(name)

    def defaultworker(self, row):
        """Perform the Step's operation on the given row.

           Inheriting classes should implement this method.
        """
        pass


class SourceStep(Step):
    """A Step that iterates over a data source and gives each row to the 
       next step. The start method must be called.
    """

    def __init__(self, source, next=None, name=None):
        """Arguments:
           - source: The data source. Must be iterable.
           - next: The default next step to use. This should be 1) an instance
             of a Step, 2) the name of a Step, or 3) None.
             If if is a name, the next step will be looked up dynamically
             each time. If it is None, no default step will exist and rows
             will not be passed on. Default: None
           - name: A name for the Step instance. This is used when another
             Step (implicitly or explicitly) passes on rows. If two instanes
             have the same name, the name is mapped to the instance that was
             created the latest. Default: None
        """
        Step.__init__(self, None, next, name)
        self.source = source

    def start(self):
        """Start the iteration of the source's rows and pass them on."""
        for row in self.source:
            self.process(row)


class MappingStep(Step):
    """A Step that applies functions to attributes in rows."""

    def __init__(self, targets, requiretargets=True, next=None, name=None):
        """Argument:
           - targets: A sequence of (name, function) pairs. For each element,
             row[name] is set to function(row[name]) for each row given to the
             step.
           - requiretargets: A flag that decides if a KeyError should be raised
             if a name from targets does not exist in a row. If True, a 
             KeyError is raised, if False the missing attribute is ignored and
             not set. Default: True
           - next: The default next step to use. This should be 1) an instance
             of a Step, 2) the name of a Step, or 3) None.
             If if is a name, the next step will be looked up dynamically
             each time. If it is None, no default step will exist and rows
             will not be passed on. Default: None
           - name: A name for the Step instance. This is used when another
             Step (implicitly or explicitly) passes on rows. If two instanes
             have the same name, the name is mapped to the instance that was
             created the latest. Default: None
        """
        Step.__init__(self, None, next, name)
        self.targets = targets
        self.requiretargets = requiretargets

    def defaultworker(self, row):
        for (element, function) in self.targets:
            if element in row:
                    row[element] = function(row[element])
            elif self.requiretargets:
                raise KeyError, "%s not found in row" % (element,)


class ValueMappingStep(Step):
    """A Step that Maps values to other values (e.g., DK -> Denmark)"""

    def __init__(self, outputatt, inputatt, mapping, requireinput=True, 
                 defaultvalue=None, next=None, name=None):
        """Arguments:
           - outputatt: The attribute to write the mapped value to in each row.
           - inputatt: The attribute to map.
           - mapping: A dict with the mapping itself.
           - requireinput: A flag that decides if a KeyError should be raised
             if inputatt does not exist in a given row. If True, a KeyError
             will be raised when the attriubte is missing. If False, a 
             the outputatt will be set to defaultvalue. Default: True
           - defaultvalue: The default value to use when the mapping cannot be 
             done. Default: None
           - next: The default next step to use. This should be 1) an instance
             of a Step, 2) the name of a Step, or 3) None.
             If if is a name, the next step will be looked up dynamically
             each time. If it is None, no default step will exist and rows
             will not be passed on. Default: None
           - name: A name for the Step instance. This is used when another
             Step (implicitly or explicitly) passes on rows. If two instanes
             have the same name, the name is mapped to the instance that was
             created the latest. Default: None
        """
        Step.__init__(self, None, next, name)
        self.outputatt = outputatt
        self.inputatt = inputatt
        self.mapping = mapping
        self.defaultvalue = defaultvalue
        self.requireinput = requireinput

    def defaultworker(self, row):
        if self.inputatt in row:
            row[self.outputatt] = self.mapping.get(row[self.inputatt],
                                                   self.defaultvalue)
        elif not self.requireinput:
            row[self.attribute] = self.defaultvalue
        else:
            raise KeyError, "%s not found in row" % (self.attribute,)


class PrintStep(Step):
    """A Step that prints each given row."""

    def __init__(self, next=None, name=None):
        """Arguments:
           - next: The default next step to use. This should be 1) an instance
             of a Step, 2) the name of a Step, or 3) None.
             If if is a name, the next step will be looked up dynamically
             each time. If it is None, no default step will exist and rows
             will not be passed on. Default: None
           - name: A name for the Step instance. This is used when another
             Step (implicitly or explicitly) passes on rows. If two instanes
             have the same name, the name is mapped to the instance that was
             created the latest. Default: None
        """
        Step.__init__(self, None, next, name)

    def defaultworker(self, row):
        print(row)


class DimensionStep(Step):
    """A Step that performs ensure(row) on a given dimension for each row."""

    def __init__(self, dimension, keyfield=None, next=None, name=None):
        """Arguments:
           - dimension: the Dimension object to call ensure on.
           - keyfield: the name of the attribute that in each row is set to 
             hold the key value for the dimension member
           - next: The default next step to use. This should be 1) an instance
             of a Step, 2) the name of a Step, or 3) None.
             If if is a name, the next step will be looked up dynamically
             each time. If it is None, no default step will exist and rows
             will not be passed on. Default: None
           - name: A name for the Step instance. This is used when another
             Step (implicitly or explicitly) passes on rows. If two instanes
             have the same name, the name is mapped to the instance that was
             created the latest. Default: None
        """
        Step.__init__(self, None, next, name)
        self.dimension = dimension
        self.keyfield = keyfield

    def defaultworker(self, row):
        key = self.dimension.ensure(row)
        if self.keyfield is not None:
            row[self.keyfield] = key


class SCDimensionStep(Step):
    """A Step that performs scdensure(row) on a given dimension for each row."""

    def __init__(self, dimension, next=None, name=None):
        """Arguments:
           - dimension: the Dimension object to call ensure on.
           - keyfield: the name of the attribute that in each row is set to 
             hold the key value for the dimension member
           - next: The default next step to use. This should be 1) an instance
             of a Step, 2) the name of a Step, or 3) None.
             If if is a name, the next step will be looked up dynamically
             each time. If it is None, no default step will exist and rows
             will not be passed on. Default: None
           - name: A name for the Step instance. This is used when another
             Step (implicitly or explicitly) passes on rows. If two instanes
             have the same name, the name is mapped to the instance that was
             created the latest. Default: None
        """
        Step.__init__(self, None, next, name)
        self.dimension = dimension

    def defaultworker(self, row):
        self.dimension.scdensure(row)


class RenamingFromToStep(Step):
    # Performs renamings of attributes in rows.
    def __init__(self, renaming, next=None, name=None):
        """Arguments:
           - name: A name for the Step instance. This is used when another
             Step (implicitly or explicitly) passes on rows. If two instanes
             have the same name, the name is mapped to the instance that was
             created the latest. Default: None
           - renaming: A dict with pairs (oldname, newname) which will
             by used by pygrametl.renamefromto to do the renaming
           - next: The default next step to use. This should be 1) an instance
             of a Step, 2) the name of a Step, or 3) None.
             If if is a name, the next step will be looked up dynamically
             each time. If it is None, no default step will exist and rows
             will not be passed on. Default: None
           - name: A name for the Step instance. This is used when another
             Step (implicitly or explicitly) passes on rows. If two instanes
             have the same name, the name is mapped to the instance that was
             created the latest. Default: None
        """
        Step.__init__(self, None, next, name)
        self.renaming = renaming

    def defaultworker(self, row):
        pygrametl.renamefromto(row, self.renaming)

RenamingStep = RenamingFromToStep # for backwards compat.


class RenamingToFromStep(RenamingFromToStep):
    def defaultworker(self, row):
        pygrametl.renametofrom(row, self.renaming)


class GarbageStep(Step):
    """ A Step that does nothing. Rows are neither modified nor passed on."""

    def __init__(self, name=None):
        """Argument:
            - name: A name for the Step instance. This is used when another
             Step (implicitly or explicitly) passes on rows. If two instanes
             have the same name, the name is mapped to the instance that was
             created the latest. Default: None
        """
        Step.__init__(self, None, None, name)

    def process(self, row):
        return


class ConditionalStep(Step):
    """A Step that redirects rows based on a condition."""

    def __init__(self, condition, whentrue, whenfalse=None, name=None):
        """Arguments:
           - condition: A function f(row) that is evaluated for each row.
           - whentrue: The next step to use if the condition evaluates to a
             true value. This argument  should be 1) an instance of a Step, 
             2) the name of a Step, or 3) None.
             If if is a name, the next step will be looked up dynamically
             each time. If it is None, no default step will exist and rows
             will not be passed on.
           - whenfalse: The Step that rows are sent to when the condition 
             evaluates to a false value. If None, the rows are silently 
             discarded. Default=None
           - name: A name for the Step instance. This is used when another
             Step (implicitly or explicitly) passes on rows. If two instanes
             have the same name, the name is mapped to the instance that was
             created the latest. Default: None
        """
        Step.__init__(self, None, whentrue, name)
        self.whenfalse = whenfalse
        self.condition = condition
        self.__nowhere = GarbageStep()

    def defaultworker(self, row):
        if not self.condition(row):
            if self.whenfalse is None:
                self._redirect(self.__nowhere)
            else:
                self._redirect(self.whenfalse)
        # else process will pass on the row to self.next (the whentrue step)


class CopyStep(Step):
    """A Step that copies each row and passes on the copy and the original"""

    def __init__(self, originaldest, copydest, deepcopy=False, name=None):
        """Arguments:
           - originaldest: The Step each given row is passed on to.
             This argument  should be 1) an instance of a Step, 
             2) the name of a Step, or 3) None.
             If if is a name, the next step will be looked up dynamically
             each time. If it is None, no default step will exist and rows
             will not be passed on.
           - copydest: The Step a copy of each given row is passed on to.
             This argument can be 1) an instance of a Step or 2) the name
             of a step.
           - name: A name for the Step instance. This is used when another
             Step (implicitly or explicitly) passes on rows. If two instanes
             have the same name, the name is mapped to the instance that was
             created the latest. Default: None
           - deepcopy: Decides if the copy should be deep or not. 
             Default: False
        """
        Step.__init__(self, None, originaldest, name)
        if copydest is None:
            raise ValueError, 'copydest is None'
        self.copydest = copydest
        import copy
        if deepcopy:
            self.copyfunc = copy.deepcopy
        else:
            self.copyfunc = copy.copy

    def defaultworker(self, row):
        copy = self.copyfunc(row)
        self._inject(copy, self.copydest)
        # process will pass on row to originaldest = self.next

# For aggregations. Experimental.


class AggregatedRow(dict):
    pass


class AggregatingStep(Step):
    def __init__(self, aggregator=None, finalizer=None, next=None, name=None):
        Step.__init__(self, aggregator, next, name)
        self.finalizer = finalizer or self.defaultfinalizer

    def process(self, row):
        if isinstance(row, AggregatedRow):
            self.finalizer(row)
            if self.next is not None:
                Step._inject(self, row, self.next)
        else:
            self.worker(row)

    def defaultworker(self, row):
        pass

    def defaultfinalizer(self, row):
        pass


class SumAggregator(AggregatingStep):
    def __init__(self, field, next=None, name=None):
        AggregatingStep.__init__(self, None, None, next, name)
        self.sum = 0
        self.field = field

    def defaultworker(self, row):
        self.sum += row[self.field]

    def defaultfinalizer(self, row):
        row[self.field] = self.sum
        self.sum = 0


class AvgAggregator(AggregatingStep):
    def __init__(self, field, next=None, name=None):
        AggregatingStep.__init__(self, None, None, next, name)
        self.sum = 0
        self.cnt = 0
        self.field = field

    def defaultworker(self, row):
        self.sum += row[self.field]
        self.cnt += 1

    def defaultfinalizer(self, row):
        if self.cnt > 0:
            row[self.field] = self.sum / float(self.cnt)
        else:
            row[self.field] = 0

        self.sum = 0
        self.cnt = 0


class MaxAggregator(AggregatingStep):
    def __init__(self, field, next=None, name=None):
        AggregatingStep.__init__(self, None, None, next, name)
        self.max = None
        self.field = field

    def defaultworker(self, row):
        if self.max is None or row[self.field] > self.max:
            self.max = row[self.field]

    def defaultfinalizer(self, row):
        row[self.field] = self.max
        self.max = None


class MinAggregator(AggregatingStep):
    def __init__(self, field, next=None, name=None):
        AggregatingStep.__init__(self, None, None, next, name)
        self.min = None
        self.field = field

    def defaultworker(self, row):
        if self.min is None or row[self.field] < self.min:
            self.min = row[self.field]

    def defaultfinalizer(self, row):
        row[self.field] = self.min
        self.min = None